/*

gbmrx.c - Generalised Bitmap Module (GBM) REXX adapter library

Author  : Heiko Nitzsche
License : Public Domain
Requires: GBM.DLL version 1.35 or higher.

History
-------
15-May-2006: Inital version (1.0)

*/

#define  INCL_DOS
#define  INCL_REXXSAA
#include <os2.h>
#include <rexxsaa.h>

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdarg.h>
#include <io.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "gbm.h"

#include "rexxhlp.h"


/*********************************************************************/
/* GBMRX.DLL version                                                 */
/*********************************************************************/
#define GBM_VERSION_REXX     "1.00"

/*********************************************************************/
/* Numeric Error Return Strings                                      */
/*********************************************************************/
#define  NOERROR_RET         "0"              /* No error whatsoever */

/*********************************************************************/
/* Alpha Numeric Return Strings                                      */
/*********************************************************************/
#define  ERROR_RETSTR           "ERROR"  /* Unspecific error occured */
#define  ERROR_MEM              "ERROR: Out of memory"
#define  ERROR_TOOMANY_TYPES    "ERROR: Too many filetypes found"
#define  ERROR_REXX_SET         "ERROR: Cannot set REXX variable"
#define  ERROR_FILETYPE         "ERROR: Invalid file type"
#define  ERROR_FILEOPEN         "ERROR: Cannot open file"
#define  ERROR_FILECREATE       "ERROR: Cannot create file"
#define  ERROR_BPP_UNSUPP       "ERROR: Colour depth is not supported"
#define  ERROR_PAGEQUERY_UNSUPP "ERROR: Query requires GBM.DLL >=1.35"

/*********************************************************************/
/* Numeric Return calls                                              */
/*********************************************************************/
#define  INVALID_ROUTINE 40            /* Raise Rexx error           */
#define  VALID_ROUTINE    0            /* Successful completion      */


/*********************************************************************/
/*  Declare all exported functions as REXX functions.                */
/*********************************************************************/
RexxFunctionHandler GBM_LoadFuncs;
RexxFunctionHandler GBM_DropFuncs;

RexxFunctionHandler GBM_Version;
RexxFunctionHandler GBM_VersionRexx;

RexxFunctionHandler GBM_Types;
RexxFunctionHandler GBM_IsBppSupported;

RexxFunctionHandler GBM_FilePages;
RexxFunctionHandler GBM_FileType;
RexxFunctionHandler GBM_FileHeader;
RexxFunctionHandler GBM_FilePalette;
RexxFunctionHandler GBM_FileData;

RexxFunctionHandler GBM_FileWrite;


/*********************************************************************/
/* RxFncTable                                                        */
/*   Array of names of the REXXUTIL functions.                       */
/*   This list is used for registration and deregistration.          */
/*********************************************************************/

static PSZ RxFncTable[] =
{
  "GBM_LoadFuncs",
  "GBM_DropFuncs",
  "GBM_Version",
  "GBM_VersionRexx",
  "GBM_Types",
  "GBM_IsBppSupported",
  "GBM_FilePages",
  "GBM_FileType",
  "GBM_FileHeader",
  "GBM_FilePalette",
  "GBM_FileData",
  "GBM_FileWrite"
};


/*********************************************************************/
/* Some useful macros                                                */
/*********************************************************************/
#define BUILDRXSTRING(t, s) { \
  strcpy((t)->strptr,(s));\
  (t)->strlength = strlen((s)); \
}


/*********************************************************************/
/* Helper functions                                                  */
/*********************************************************************/
static char gbm_error_msg[61] = { 0 };

static const char * formatGbmErrorMessage(const GBM_ERR errorCode)
{
  sprintf(gbm_error_msg, "ERROR: %.53s", Gbm_err(errorCode));
  return gbm_error_msg;
}

/*********************************************************************/

static BOOL isEqualLowerCase(const char *s1, const char *s2, int n)
{
  for ( ; n--; s1++, s2++ )
  {
    if ( tolower(*s1) != tolower(*s2) )
    {
      return FALSE;
    }
  }
  return TRUE;
}

/*********************************************************************/

static const char *findWord(const char *str, const char *substr)
{
         char * buf, *s;
         char   strbuf[51];
         int    len            = strlen(substr);
   const int    strlength      = strlen(str);
   const BOOL   needsDynBuffer = (strlength > (sizeof(strbuf)-1));

   /* only use dynamic buffer if needed to speedup a bit */
   if (needsDynBuffer)
   {
     buf = (char *) malloc(strlength + 1);
     if (buf == NULL)
     {
       return NULL;
     }
   }
   else
   {
     buf = strbuf;
   }

   for ( s  = strtok(strcpy(buf, str), " \t,");
         s != NULL;
         s  = strtok(NULL, " \t,") )
   {
     if ( isEqualLowerCase(s, substr, len) && s[len] == '\0' )
     {
       int inx = s - buf;
       if (needsDynBuffer)
       {
         free(buf);
       }
       return str + inx;
     }
   }
   if (needsDynBuffer)
   {
     free(buf);
   }
   return NULL;
}

/*********************************************************************/

static int lookupGbmFileType(const char * extension)
{
  int numFileTypes = -1;

  if (Gbm_query_n_filetypes(&numFileTypes) == GBM_ERR_OK)
  {
    int     ft;
    GBMFT   gbmft;

    for (ft = 0; ft < numFileTypes; ft++)
    {
      if (Gbm_query_filetype(ft, &gbmft) == GBM_ERR_OK)
      {
        if (findWord(gbmft.extensions, extension) != NULL)
        {
          return ft;
        }
      }
    }
  }
  return -1;
}

/*********************************************************************/

/* Depending on GBM.DLL version the number of pages can be retrieved (versions > 1.35) */
/* or the functionality does not yet exist.                                            */
/*                                                                                     */
/* Dynamically link the specific function to support also older versions of GBM.DLL.   */

BOOL getNumberOfPages(const char * fileName, const int fd, const int ft, int * numPages)
{
   HMODULE hmod;
   PFN     functionAddr = NULL;
   APIRET  rc = 0;

   /* check version first */
   if (Gbm_version() < 135)
   {
      return FALSE;
   }

   /* now dynamically link GBM.DLL */
   rc = DosLoadModule("", 0, "GBM.DLL", &hmod);
   if (rc)
   {
      return FALSE;
   }

   /* lookup gbm_read_imgcount() */
   rc = DosQueryProcAddr(hmod, 0L, "gbm_read_imgcount", &functionAddr);
   if (rc)
   {
      DosFreeModule(hmod);
      return FALSE;
   }

   /* call gbm_read_imgcount(const char *fn, int fd, int ft, int *pimgcnt) */
   if (functionAddr(fileName, fd, ft, numPages) != GBM_ERR_OK)
   {
      DosFreeModule(hmod);
      return FALSE;
   }

   DosFreeModule(hmod);

   return TRUE;
}


/*********************************************************************/
/* Function:  GBM_LoadFuncs                                          */
/*                                                                   */
/* Syntax:    call GBM_LoadFuncs                                     */
/*                                                                   */
/* Params:    none                                                   */
/*                                                                   */
/* Return:    null string                                            */
/*********************************************************************/

ULONG GBM_LoadFuncs(PUCHAR name, ULONG numargs, RXSTRING args[],
                    PSZ queuename, RXSTRING *retstr)
{
  /* prevent compiler warnings */
  args      = args;
  name      = name;
  queuename = queuename;

  retstr->strlength = 0; /* return a null string result*/

  if (numargs != 0)
  {
    return INVALID_ROUTINE;
  }

  RegisterRexxFunctions("GBMRX",
                        RxFncTable,
                        sizeof(RxFncTable)/sizeof(PSZ));

  return VALID_ROUTINE;  /* no error on call */
}

/*********************************************************************/
/* Function:  GBM_DropFuncs                                          */
/*                                                                   */
/* Syntax:    call GBM_DropFuncs                                     */
/*                                                                   */
/* Params:    none                                                   */
/*                                                                   */
/* Return:    null string                                            */
/*********************************************************************/

ULONG GBM_DropFuncs(PUCHAR name, ULONG numargs, RXSTRING args[],
                    PSZ queuename, RXSTRING *retstr)
{
  /* prevent compiler warnings */
  args      = args;
  name      = name;
  queuename = queuename;

  retstr->strlength = 0;           /* return a null string result*/

  if (numargs != 0)                /* no arguments for this      */
  {
    return INVALID_ROUTINE;
  }

  UnregisterRexxFunctions(RxFncTable,
                          sizeof(RxFncTable)/sizeof(PSZ));

  return VALID_ROUTINE;  /* no error on call */
}


/*********************************************************************/
/* Function:  GBM_Version                                            */
/*                                                                   */
/* Syntax:    version = GBM_Version()                                */
/*                                                                   */
/* Return:    Version number: "major.minor"                          */
/*********************************************************************/
#define GET_GBM_VERSION(version)  ((float)version/100.0)

ULONG GBM_Version(PUCHAR name, ULONG numargs, RXSTRING args[],
                  PSZ queuename, RXSTRING *retstr)
{
  int version;

  /* prevent compiler warnings */
  args      = args;
  name      = name;
  queuename = queuename;

  retstr->strlength = 0;           /* return a null string result*/

  if (numargs != 0)                /* no arguments for this      */
  {
    return INVALID_ROUTINE;
  }

  Gbm_init();
  version = Gbm_version();
  Gbm_deinit();

  /* build return string ("major.minor")*/
  sprintf(retstr->strptr, "%.2f", GET_GBM_VERSION(version));
  retstr->strlength = strlen(retstr->strptr);

  return VALID_ROUTINE;  /* no error on call */
}


/*********************************************************************/
/* Function:  GBM_VersionRexx                                        */
/*                                                                   */
/* Syntax:    version = GBM_VersionRexx()                            */
/*                                                                   */
/* Return:    Version number: "major.minor"                          */
/*********************************************************************/

ULONG GBM_VersionRexx(PUCHAR name, ULONG numargs, RXSTRING args[],
                      PSZ queuename, RXSTRING *retstr)
{
  /* prevent compiler warnings */
  args      = args;
  name      = name;
  queuename = queuename;

  retstr->strlength = 0;           /* return a null string result*/

  if (numargs != 0)                /* no arguments for this      */
  {
    return INVALID_ROUTINE;
  }

  /* build return string ("major.minor")*/
  BUILDRXSTRING(retstr, GBM_VERSION_REXX);

  return VALID_ROUTINE;  /* no error on call */
}


/*******************************************************************/
/* Function:  GBM_Types                                            */
/*                                                                 */
/* Syntax:    rc = GBM_Types(stem)                                 */
/*                                                                 */
/* Params:    stem out - bitmap format information consisting of   */
/*                       stem.0           : number of info blocks  */
/*                       stem.X.extensions: bitmap extensions      */
/*                       stem.X.shortname : short description      */
/*                       stem.X.longname  : long description       */
/*                                                                 */
/*                       X: Ranging from 1 to stem.0               */
/*                                                                 */
/* Return:   "0", an error message or "ERROR" for an unknown error */
/*******************************************************************/

ULONG GBM_Types(PUCHAR name, ULONG numargs, RXSTRING args[],
                PSZ queuename, RXSTRING *retstr)
{
  int       numFileTypes = 0;
  char    * infoStem     = NULL;
  char    * infoStemSep  = "";
  GBM_ERR   rc           = GBM_ERR_NOT_SUPP;

  /* prevent compiler warnings */
  args      = args;
  name      = name;
  queuename = queuename;

  BUILDRXSTRING(retstr, ERROR_RETSTR);

  if (numargs != 1)
  {
    return INVALID_ROUTINE;
  }

  /* extract arguments */
  infoStem    = args[0].strptr;
  infoStemSep = (infoStem[strlen(infoStem)-1] != '.') ? "." : "";

  Gbm_init();

  /* build return string */
  rc = Gbm_query_n_filetypes(&numFileTypes);
  if (rc == GBM_ERR_OK)
  {
    int     ft;
    GBMFT   gbmft;
    char  * stemBuf    = NULL;
    char    varBuf[21] = { 0 };

    /* prepare the stem buffer */
    const int maxStemLength = strlen(infoStem) + strlen(infoStemSep);

    /* we support only 999 file types which should be enough */
    if (numFileTypes > 999)
    {
      Gbm_deinit();
      BUILDRXSTRING(retstr, ERROR_TOOMANY_TYPES);
      return VALID_ROUTINE;
    }

    /* allocate the stem name buffer + 3 digits for the index + info name (12) */
    stemBuf = (char *) malloc(maxStemLength + 3 /* stem index */ + 12 /* info name */ + 1);
    if (stemBuf == NULL)
    {
      Gbm_deinit();
      BUILDRXSTRING(retstr, ERROR_MEM);
      return VALID_ROUTINE;
    }

    /* now set the types to the stems */
    for (ft = 0; ft < numFileTypes; ft++)
    {
       rc = Gbm_query_filetype(ft, &gbmft);
      if (rc != GBM_ERR_OK)
      {
        BUILDRXSTRING(retstr, formatGbmErrorMessage(rc));
        Gbm_deinit();
        return VALID_ROUTINE;
      }

      /* set to the stem */
      sprintf(stemBuf, "%s%s%i.extensions", infoStem, infoStemSep, ft + 1);
      if (! SetRexxVariable(stemBuf, gbmft.extensions))
      {
        free(stemBuf);
        Gbm_deinit();
        BUILDRXSTRING(retstr, ERROR_REXX_SET);
        return VALID_ROUTINE;
      }
      sprintf(stemBuf, "%s%s%i.shortname", infoStem, infoStemSep, ft + 1);
      if (! SetRexxVariable(stemBuf, gbmft.short_name))
      {
        free(stemBuf);
        Gbm_deinit();
        BUILDRXSTRING(retstr, ERROR_REXX_SET);
        return VALID_ROUTINE;
      }
      sprintf(stemBuf, "%s%s%i.longname", infoStem, infoStemSep, ft + 1);
      if (! SetRexxVariable(stemBuf, gbmft.long_name))
      {
        free(stemBuf);
        Gbm_deinit();
        BUILDRXSTRING(retstr, ERROR_REXX_SET);
        return VALID_ROUTINE;
      }
    }

    /* set the number of types into stem.0 */
    sprintf(varBuf, "%i", numFileTypes);
    sprintf(stemBuf, "%s%s0", infoStem, infoStemSep);
    if (! SetRexxVariable(stemBuf, varBuf))
    {
      free(stemBuf);
      Gbm_deinit();
      BUILDRXSTRING(retstr, ERROR_REXX_SET);
      return VALID_ROUTINE;
    }

    free(stemBuf);
    Gbm_deinit();

    BUILDRXSTRING(retstr, NOERROR_RET);
    return VALID_ROUTINE;
  }

  BUILDRXSTRING(retstr, formatGbmErrorMessage(rc));

  Gbm_deinit();
  return VALID_ROUTINE;
}


/************************************************************************/
/* Function:  GBM_IsBppSupported                                        */
/*                                                                      */
/* Syntax:    sup = GBM_IsBppSupported(fileExtension, bpp, rw)          */
/*                                                                      */
/* Params:    string  in - the filetype extension reported by GBM_Types */
/*                                                                      */
/*            string  in - bpp (bits per pixel) to test                 */
/*                                                                      */
/*            string  in - "r" for testing for read support             */
/*                         "w" for testing for read support             */
/*                                                                      */
/* Return:   "0" (unsupported) or "1" (supported)                       */
/************************************************************************/

ULONG GBM_IsBppSupported(PUCHAR name, ULONG numargs, RXSTRING args[],
                         PSZ queuename, RXSTRING *retstr)
{
  char    * extension    = NULL;
  char    * readwrite    = NULL;
  int       filetype     = -1;
  int       bpp          = -1;
  GBMFT     gbmft        = { 0 };
  GBM_ERR   rc           = GBM_ERR_NOT_SUPP;

  /* prevent compiler warnings */
  args      = args;
  name      = name;
  queuename = queuename;

  BUILDRXSTRING(retstr, "0");

  if (numargs != 3)
  {
    return INVALID_ROUTINE;
  }

  /* extract arguments */
  extension = args[0].strptr;
  readwrite = args[2].strptr;

  if (sscanf(args[1].strptr, "%i", &bpp) != 1)
  {
    return VALID_ROUTINE;
  }

  Gbm_init();

  filetype = lookupGbmFileType(extension);
  if (filetype < 0)
  {
    Gbm_deinit();
    return VALID_ROUTINE;
  }

  /* check if the output format supports the requested color depth */
  rc = Gbm_query_filetype(filetype, &gbmft);
  if (rc != GBM_ERR_OK)
  {
    Gbm_deinit();
    return VALID_ROUTINE;
  }

  Gbm_deinit();

  if (isEqualLowerCase(readwrite, "r", 1))
  {
    int flag = 0;
    switch(bpp)
    {
      case 64: flag = GBM_FT_R64; break;
      case 48: flag = GBM_FT_R48; break;
      case 32: flag = GBM_FT_R32; break;
      case 24: flag = GBM_FT_R24; break;
      case  8: flag = GBM_FT_R8;  break;
      case  4: flag = GBM_FT_R4;  break;
      case  1: flag = GBM_FT_R1;  break;
      default: flag = 0;          break;
    }
    if ( (gbmft.flags & flag) == 0 )
    {
      return VALID_ROUTINE;
    }
    BUILDRXSTRING(retstr, "1");
    return VALID_ROUTINE;
  }

  if (isEqualLowerCase(readwrite, "w", 1))
  {
    int flag = 0;
    switch(bpp)
    {
      case 64: flag = GBM_FT_W64; break;
      case 48: flag = GBM_FT_W48; break;
      case 32: flag = GBM_FT_W32; break;
      case 24: flag = GBM_FT_W24; break;
      case  8: flag = GBM_FT_W8;  break;
      case  4: flag = GBM_FT_W4;  break;
      case  1: flag = GBM_FT_W1;  break;
      default: flag = 0;          break;
    }
    if ( (gbmft.flags & flag) == 0 )
    {
      return VALID_ROUTINE;
    }
    BUILDRXSTRING(retstr, "1");
    return VALID_ROUTINE;
  }

  return VALID_ROUTINE;  /* no error on call */
}


/*******************************************************************/
/* Function:  GBM_FileType                                         */
/*                                                                 */
/* Syntax:    rc = GBM_FileType(filename, stem)                    */
/*                                                                 */
/* Params:    string  in - filename                                */
/*                                                                 */
/*            stem   out - bitmap format information consisting of */
/*                         stem.extensions: bitmap extensions      */
/*                         stem.shortname : short description      */
/*                         stem.longname  : long description       */
/*                                                                 */
/* Return:   "0", an error message or "ERROR" for an unknown error */
/*                                                                 */
/* Note: This function does only check the codec database but not  */
/*       the file content. Use GBM_FileHeader for this.            */
/*                                                                 */
/*******************************************************************/

ULONG GBM_FileType(PUCHAR name, ULONG numargs, RXSTRING args[],
                   PSZ queuename, RXSTRING *retstr)
{
  char    * filename     = NULL;
  char    * infoStem     = NULL;
  char    * infoStemSep  = NULL;
  char    * infoStemFull = NULL;
  int       filetype     = -1;
  GBMFT     gbmft        = { 0 };
  GBM_ERR   rc           = GBM_ERR_NOT_SUPP;

  /* prevent compiler warnings */
  args      = args;
  name      = name;
  queuename = queuename;

  BUILDRXSTRING(retstr, ERROR_RETSTR);

  if (numargs != 2)
  {
    return INVALID_ROUTINE;
  }

  /* extract arguments */
  filename    = args[0].strptr;
  infoStem    = args[1].strptr;
  infoStemSep = (infoStem[strlen(infoStem)-1] != '.') ? "." : "";

  Gbm_init();

  /* query filetype from file extension */
  rc = Gbm_guess_filetype(filename, &filetype);
  if (rc != GBM_ERR_OK)
  {
    BUILDRXSTRING(retstr, formatGbmErrorMessage(rc));
    Gbm_deinit();
    return VALID_ROUTINE;
  }
  rc = Gbm_query_filetype(filetype, &gbmft);
  if (rc != GBM_ERR_OK)
  {
    BUILDRXSTRING(retstr, formatGbmErrorMessage(rc));
    Gbm_deinit();
    return VALID_ROUTINE;
  }
  Gbm_deinit();

  /* allocate the stem name buffer + info name (12) */
  infoStemFull = (char *) malloc(strlen(infoStem) + strlen(infoStemSep) + 12 /* info name */ + 1);
  if (infoStemFull == NULL)
  {
    BUILDRXSTRING(retstr, ERROR_MEM);
    return VALID_ROUTINE;
  }

  /* set the values into the stems */
  sprintf(infoStemFull, "%s%sextensions", infoStem, infoStemSep);
  if (! SetRexxVariable(infoStemFull, gbmft.extensions))
  {
    free(infoStemFull);
    BUILDRXSTRING(retstr, ERROR_REXX_SET);
    return VALID_ROUTINE;
  }
  sprintf(infoStemFull, "%s%sshortname", infoStem, infoStemSep);
  if (! SetRexxVariable(infoStemFull, gbmft.short_name))
  {
    free(infoStemFull);
    BUILDRXSTRING(retstr, ERROR_REXX_SET);
    return VALID_ROUTINE;
  }
  sprintf(infoStemFull, "%s%slongname", infoStem, infoStemSep);
  if (! SetRexxVariable(infoStemFull, gbmft.long_name))
  {
    free(infoStemFull);
    BUILDRXSTRING(retstr, ERROR_REXX_SET);
    return VALID_ROUTINE;
  }
  free(infoStemFull);

  /* build return string */
  BUILDRXSTRING(retstr, NOERROR_RET);
  return VALID_ROUTINE;  /* no error on call */
}


/*********************************************************************/
/* Function:  GBM_FilePages                                          */
/*                                                                   */
/* Syntax:    rc = GBM_FilePages(filename [, fileExtension])         */
/*                                                                   */
/* Params:    string  in - filename                                  */
/*                                                                   */
/*            string  in - the optional filetype extension reported  */
/*                         by GBM_Types to override autodetection    */
/*                         based on the file extension.              */
/*                                                                   */
/* Return:   "0", an error message or "ERROR" for an unknown error   */
/*********************************************************************/

ULONG GBM_FilePages(PUCHAR name, ULONG numargs, RXSTRING args[],
                    PSZ queuename, RXSTRING *retstr)
{
  int       fd       = -1;
  int       imgCount = 0;
  char    * filename = NULL;
  int       filetype = -1;
  GBM_ERR   rc       = GBM_ERR_NOT_SUPP;

  /* prevent compiler warnings */
  args      = args;
  name      = name;
  queuename = queuename;

  BUILDRXSTRING(retstr, ERROR_RETSTR);

  if ((numargs < 1) || (numargs > 2))
  {
    return INVALID_ROUTINE;
  }

  Gbm_init();

  filename = args[0].strptr;
  if (numargs == 2)
  {
    filetype = lookupGbmFileType(args[1].strptr);
    if (filetype < 0)
    {
      Gbm_deinit();
      BUILDRXSTRING(retstr, ERROR_FILETYPE);
      return VALID_ROUTINE;
    }
  }

  /* open file */
  fd = Gbm_io_open(filename, GBM_O_RDONLY);
  if (fd != -1)
  {
    /* check if the client has specified no file type, use autodetection */
    if (filetype == -1)
    {
      /* query filetype from file extension */
      rc = Gbm_guess_filetype(filename, &filetype);
      if (rc != GBM_ERR_OK)
      {
        BUILDRXSTRING(retstr, formatGbmErrorMessage(rc));
        Gbm_io_close(fd);
        Gbm_deinit();
        return VALID_ROUTINE;
      }
    }

    /* get image count */
    if (! getNumberOfPages(filename, fd, filetype, &imgCount))
    {
      BUILDRXSTRING(retstr, ERROR_PAGEQUERY_UNSUPP);
      Gbm_io_close(fd);
      Gbm_deinit();
      return VALID_ROUTINE;
    }

    Gbm_io_close(fd);
    Gbm_deinit();

    /* build return string */
    sprintf(retstr->strptr, "%i", imgCount);
    retstr->strlength = strlen(retstr->strptr);

    return VALID_ROUTINE;  /* no error on call */
  }

  BUILDRXSTRING(retstr, ERROR_FILEOPEN);

  Gbm_deinit();
  return VALID_ROUTINE;
}


/*****************************************************************************/
/* Function:  GBM_FileHeader                                                 */
/*                                                                           */
/* Syntax:    rc = GBM_FileHeader(filename, options, stem [, fileExtension]) */
/*                                                                           */
/* Params:    string  in - filename                                          */
/*                                                                           */
/*            string  in - comma separated options                           */
/*                         (see GBM format documentation)                    */
/*                                                                           */
/*            stem   out - header information consisting of                  */
/*                         stem.width : bitmap width                         */
/*                         stem.height: bitmap height                        */
/*                         stem.bpp   : bitmap colour depth (bpp)            */
/*                                                                           */
/*            string  in - the optional filetype extension reported          */
/*                         by GBM_Types to override autodetection            */
/*                         based on the file extension.                      */
/*                                                                           */
/* Return:   "0", an error message or "ERROR" for an unknown error           */
/*****************************************************************************/

ULONG GBM_FileHeader(PUCHAR name, ULONG numargs, RXSTRING args[],
                     PSZ queuename, RXSTRING *retstr)
{
  char    * filename    = NULL;
  char    * options     = NULL;
  char    * infoStem    = NULL;
  char    * infoStemSep = NULL;
  int       fd          = -1;
  int       filetype    = -1;
  GBM       gbm         = { 0 };
  GBM_ERR   rc          = GBM_ERR_NOT_SUPP;

  /* prevent compiler warnings */
  args      = args;
  name      = name;
  queuename = queuename;

  BUILDRXSTRING(retstr, ERROR_RETSTR);

  if ((numargs < 3) || (numargs > 4))
  {
    return INVALID_ROUTINE;
  }

  /* extract arguments */
  filename    = args[0].strptr;
  options     = args[1].strptr;
  infoStem    = args[2].strptr;
  infoStemSep = (infoStem[strlen(infoStem)-1] != '.') ? "." : "";

  Gbm_init();

  /* check if filetype override is provided */
  if (numargs > 3)
  {
    filetype = lookupGbmFileType(args[3].strptr);
    if (filetype < 0)
    {
      Gbm_deinit();
      BUILDRXSTRING(retstr, ERROR_FILETYPE);
      return VALID_ROUTINE;
    }
  }

  /* open file */
  fd = Gbm_io_open(filename, GBM_O_RDONLY);
  if (fd != -1)
  {
    char * infoStemFull = NULL;
    char   valBuf[20]   = { 0 };

    /* check if the client has specified no file type, use autodetection */
    if (filetype == -1)
    {
      /* query filetype from file extension */
      rc = Gbm_guess_filetype(filename, &filetype);
      if (rc != GBM_ERR_OK)
      {
        BUILDRXSTRING(retstr, formatGbmErrorMessage(rc));
        Gbm_io_close(fd);
        Gbm_deinit();
        return VALID_ROUTINE;
      }
    }

    /* get file header */
    rc = Gbm_read_header(filename, fd, filetype, &gbm, options);
    if (rc != GBM_ERR_OK)
    {
      BUILDRXSTRING(retstr, formatGbmErrorMessage(rc));
      Gbm_io_close(fd);
      Gbm_deinit();
      return VALID_ROUTINE;
    }

    Gbm_io_close(fd);
    Gbm_deinit();

    /* Allocate buffer for full stem name + 10 (stem subnames)) */
    infoStemFull = (char *) malloc(strlen(infoStem) + strlen(infoStemSep) + 12);
    if (infoStemFull == NULL)
    {
      BUILDRXSTRING(retstr, ERROR_MEM);
      return VALID_ROUTINE;
    }

    /* set the values into the stems */

    /* bitmap width */
    sprintf(infoStemFull, "%s%swidth", infoStem, infoStemSep);
    sprintf(valBuf, "%i", gbm.w);
    if (! SetRexxVariable(infoStemFull, valBuf))
    {
      free(infoStemFull);
      BUILDRXSTRING(retstr, ERROR_REXX_SET);
      return VALID_ROUTINE;
    }

    /* bitmap height */
    sprintf(infoStemFull, "%s%sheight", infoStem, infoStemSep);
    sprintf(valBuf, "%i", gbm.h);
    if (! SetRexxVariable(infoStemFull, valBuf))
    {
      free(infoStemFull);
      BUILDRXSTRING(retstr, ERROR_REXX_SET);
      return VALID_ROUTINE;
    }

    /* bitmap colour depth */
    sprintf(infoStemFull, "%s%sbpp", infoStem, infoStemSep);
    sprintf(valBuf, "%i", gbm.bpp);
    if (! SetRexxVariable(infoStemFull, valBuf))
    {
      free(infoStemFull);
      BUILDRXSTRING(retstr, ERROR_REXX_SET);
      return VALID_ROUTINE;
    }

    /* number of elements */
    sprintf(infoStemFull, "%s%s0", infoStem, infoStemSep);
    if (! SetRexxVariable(infoStemFull, "3"))
    {
      free(infoStemFull);
      BUILDRXSTRING(retstr, ERROR_REXX_SET);
      return VALID_ROUTINE;
    }
    free(infoStemFull);

    /* build return string */
    BUILDRXSTRING(retstr, NOERROR_RET);

    return VALID_ROUTINE;  /* no error on call */
  }

  BUILDRXSTRING(retstr, ERROR_FILEOPEN);

  Gbm_deinit();
  return VALID_ROUTINE;
}


/*****************************************************************************/
/* Function: GBM_FilePalette                                                 */
/*                                                                           */
/* Syntax:   rc = GBM_FilePalette(filename, options, stem [, fileExtension]) */
/*                                                                           */
/* Params:   string  in - filename                                           */
/*                                                                           */
/*           string  in - comma separated options                            */
/*                        (see GBM format documentation)                     */
/*                                                                           */
/*           stem   out - palette information consisting of                  */
/*                        stem.0      : number of palette entries            */
/*                        stem.X.red  : red value                            */
/*                        stem.X.green: green value                          */
/*                        stem.X.blue : blue value                           */
/*                                                                           */
/*                        X: Ranging from 1 to stem.0                        */
/*                                                                           */
/*                        Note: Not all palette entries must be              */
/*                              referenced by the bitmap data.               */
/*                                                                           */
/*           string  in - the optional filetype extension reported           */
/*                        by GBM_Types to override autodetection             */
/*                        based on the file extension.                       */
/*                                                                           */
/* Return:   "0", an error message or "ERROR" for an unknown error           */
/*****************************************************************************/

ULONG GBM_FilePalette(PUCHAR name, ULONG numargs, RXSTRING args[],
                      PSZ queuename, RXSTRING *retstr)
{
  char    * filename    = NULL;
  char    * options     = NULL;
  char    * infoStem    = NULL;
  char    * infoStemSep = NULL;
  int       fd          = -1;
  int       filetype    = -1;
  GBM       gbm         = { 0 };
  GBM_ERR   rc          = GBM_ERR_NOT_SUPP;

  /* prevent compiler warnings */
  args      = args;
  name      = name;
  queuename = queuename;

  BUILDRXSTRING(retstr, ERROR_RETSTR);

  if ((numargs < 3) || (numargs > 4))
  {
    return INVALID_ROUTINE;
  }

  /* extract arguments */
  filename    = args[0].strptr;
  options     = args[1].strptr;
  infoStem    = args[2].strptr;
  infoStemSep = (infoStem[strlen(infoStem)-1] != '.') ? "." : "";

  Gbm_init();

  /* check if filetype override is provided */
  if (numargs > 3)
  {
    filetype = lookupGbmFileType(args[3].strptr);
    if (filetype < 0)
    {
      Gbm_deinit();
      BUILDRXSTRING(retstr, ERROR_FILETYPE);
      return VALID_ROUTINE;
    }
  }

  /* open file */
  fd = Gbm_io_open(filename, GBM_O_RDONLY);
  if (fd != -1)
  {
    char * infoStemFull  = NULL;
    char   valBuf[20]    = { 0 };
    GBMRGB gbmrgb[0x100] = { 0 };

    /* check if the client has specified no file type, use autodetection */
    if (filetype == -1)
    {
      /* query filetype from file extension */
      rc = Gbm_guess_filetype(filename, &filetype);
      if (rc != GBM_ERR_OK)
      {
        BUILDRXSTRING(retstr, formatGbmErrorMessage(rc));
        Gbm_io_close(fd);
        Gbm_deinit();
        return VALID_ROUTINE;
      }
    }

    /* get file header */
    rc = Gbm_read_header(filename, fd, filetype, &gbm, options);
    if (rc != GBM_ERR_OK)
    {
      BUILDRXSTRING(retstr, formatGbmErrorMessage(rc));
      Gbm_io_close(fd);
      Gbm_deinit();
      return VALID_ROUTINE;
    }

    /* Allocate buffer for full stem name + 3 digits for the index + 10 (stem subnames)) */
    infoStemFull = (char *) malloc(strlen(infoStem) + strlen(infoStemSep) + 3 + 12);
    if (infoStemFull == NULL)
    {
      Gbm_io_close(fd);
      Gbm_deinit();
      BUILDRXSTRING(retstr, ERROR_MEM);
      return VALID_ROUTINE;
    }

    /* only colour depth up to 8bpp have a palette */
    if (gbm.bpp <= 8)
    {
      int i;
      const int palette_entries = 1 << gbm.bpp;

      /* get bitmap palette */
      rc = Gbm_read_palette(fd, filetype, &gbm, gbmrgb);
      if (rc != GBM_ERR_OK)
      {
        free(infoStemFull);
        BUILDRXSTRING(retstr, formatGbmErrorMessage(rc));
        Gbm_io_close(fd);
        Gbm_deinit();
        return VALID_ROUTINE;
      }
      Gbm_io_close(fd);
      Gbm_deinit();

      /* set the values into the stems */
      for (i = 0; i < palette_entries; i++)
      {
        /* red */
        sprintf(infoStemFull, "%s%s%i.red", infoStem, infoStemSep, i+1);
        sprintf(valBuf, "%i", gbmrgb[i].r);
        if (! SetRexxVariable(infoStemFull, valBuf))
        {
          free(infoStemFull);
          BUILDRXSTRING(retstr, ERROR_REXX_SET);
          return VALID_ROUTINE;
        }
        /* green */
        sprintf(infoStemFull, "%s%s%i.green", infoStem, infoStemSep, i+1);
        sprintf(valBuf, "%i", gbmrgb[i].g);
        if (! SetRexxVariable(infoStemFull, valBuf))
        {
          free(infoStemFull);
          BUILDRXSTRING(retstr, ERROR_REXX_SET);
          return VALID_ROUTINE;
        }
        /* blue */
        sprintf(infoStemFull, "%s%s%i.blue", infoStem, infoStemSep, i+1);
        sprintf(valBuf, "%i", gbmrgb[i].b);
        if (! SetRexxVariable(infoStemFull, valBuf))
        {
          free(infoStemFull);
          BUILDRXSTRING(retstr, ERROR_REXX_SET);
          return VALID_ROUTINE;
        }
      }

      /* number of palette entries */
      sprintf(infoStemFull, "%s%s0", infoStem, infoStemSep);
      sprintf(valBuf, "%i", palette_entries);
      if (! SetRexxVariable(infoStemFull, valBuf))
      {
        free(infoStemFull);
        BUILDRXSTRING(retstr, ERROR_REXX_SET);
        return VALID_ROUTINE;
      }
    }
    else
    {
      Gbm_io_close(fd);
      Gbm_deinit();

      /* number of elements */
      sprintf(infoStemFull, "%s%s0", infoStem, infoStemSep);
      if (! SetRexxVariable(infoStemFull, "0"))
      {
        free(infoStemFull);
        BUILDRXSTRING(retstr, ERROR_REXX_SET);
        return VALID_ROUTINE;
      }
    }

    free(infoStemFull);

    /* build return string */
    BUILDRXSTRING(retstr, NOERROR_RET);

    return VALID_ROUTINE;  /* no error on call */
  }

  BUILDRXSTRING(retstr, ERROR_FILEOPEN);

  Gbm_deinit();
  return VALID_ROUTINE;
}


/*****************************************************************************/
/* Function:  GBM_FileData                                                   */
/*                                                                           */
/* Syntax:    rc = GBM_FileData(filename, options, stem [, fileExtension])   */
/*                                                                           */
/* Params:    string  in - filename                                          */
/*                                                                           */
/*            string  in - comma separated options                           */
/*                         (see GBM format documentation)                    */
/*                                                                           */
/*            stem   out - The binary bitmap data.                           */
/*                                                                           */
/*            string  in - the optional filetype extension reported          */
/*                         by GBM_Types to override autodetection            */
/*                         based on the file extension.                      */
/*                                                                           */
/* Return:   "0", an error message or "ERROR" for an unknown error           */
/*****************************************************************************/

ULONG GBM_FileData(PUCHAR name, ULONG numargs, RXSTRING args[],
                   PSZ queuename, RXSTRING *retstr)
{
  char    * filename = NULL;
  char    * options  = NULL;
  char    * dataStem = NULL;
  int       fd       = -1;
  int       filetype = -1;
  GBM       gbm      = { 0 };
  GBM_ERR   rc       = GBM_ERR_NOT_SUPP;

  /* prevent compiler warnings */
  args      = args;
  name      = name;
  queuename = queuename;

  BUILDRXSTRING(retstr, ERROR_RETSTR);

  if ((numargs < 3) || (numargs > 4))
  {
    return INVALID_ROUTINE;
  }

  /* extract arguments */
  filename = args[0].strptr;
  options  = args[1].strptr;
  dataStem = args[2].strptr;

  Gbm_init();

  /* check if filetype override is provided */
  if (numargs > 3)
  {
    filetype = lookupGbmFileType(args[3].strptr);
    if (filetype < 0)
    {
      Gbm_deinit();
      BUILDRXSTRING(retstr, ERROR_FILETYPE);
      return VALID_ROUTINE;
    }
  }

  /* open file */
  fd = Gbm_io_open(filename, GBM_O_RDONLY);
  if (fd != -1)
  {
    int    stride  = 0;
    int    datalen = 0;
    char * data    = NULL;

    /* check if the client has specified no file type, use autodetection */
    if (filetype == -1)
    {
      /* query filetype from file extension */
      rc = Gbm_guess_filetype(filename, &filetype);
      if (rc != GBM_ERR_OK)
      {
        BUILDRXSTRING(retstr, formatGbmErrorMessage(rc));
        Gbm_io_close(fd);
        Gbm_deinit();
        return VALID_ROUTINE;
      }
    }

    /* get file header */
    rc = Gbm_read_header(filename, fd, filetype, &gbm, options);
    if (rc != GBM_ERR_OK)
    {
      BUILDRXSTRING(retstr, formatGbmErrorMessage(rc));
      Gbm_io_close(fd);
      Gbm_deinit();
      return VALID_ROUTINE;
    }

    /* Allocate a buffer for the bitmap data */
    stride  = ( ((gbm.w * gbm.bpp + 31)/32) * 4 );
    datalen = stride * gbm.h;
    data    = (char *) malloc((size_t) datalen);
    if (data == NULL)
    {
      Gbm_io_close(fd);
      Gbm_deinit();
      BUILDRXSTRING(retstr, ERROR_MEM);
      return VALID_ROUTINE;
    }

    /* Read the bitmap data */
    rc = Gbm_read_data(fd, filetype, &gbm, data);
    if (rc != GBM_ERR_OK)
    {
      free(data);
      BUILDRXSTRING(retstr, formatGbmErrorMessage(rc));
      Gbm_io_close(fd);
      Gbm_deinit();
      return VALID_ROUTINE;
    }

    Gbm_io_close(fd);
    Gbm_deinit();

    /* set the data into the stem */
    if (! SetRexxBinaryVariable(dataStem, data, datalen))
    {
      free(data);
      BUILDRXSTRING(retstr, ERROR_REXX_SET);
      return VALID_ROUTINE;
    }

    free(data);

    /* build return string */
    BUILDRXSTRING(retstr, NOERROR_RET);

    return VALID_ROUTINE;  /* no error on call */
  }

  BUILDRXSTRING(retstr, ERROR_FILEOPEN);

  Gbm_deinit();
  return VALID_ROUTINE;
}


/*******************************************************************************************/
/* Function:  GBM_FileWrite                                                                */
/*                                                                                         */
/* Syntax:    rc = GBM_FileWrite(filename, options, stem1, stem2, stem3 [, fileExtension]) */
/*                                                                                         */
/* Params:    string  in - filename                                                        */
/*                                                                                         */
/*            string  in - comma separated options                                         */
/*                         (see GBM format documentation)                                  */
/*                                                                                         */
/*            stem1   in - header information consisting of                                */
/*                         stem1.width : bitmap width                                      */
/*                         stem1.height: bitmap height                                     */
/*                         stem1.bpp   : bitmap colour depth (bpp)                         */
/*                                                                                         */
/*            stem2   in - palette information consisting of                               */
/*                         stem2.0      : number of palette entries                        */
/*                         stem2.X.red  : red value                                        */
/*                         stem2.X.green: green value                                      */
/*                         stem2.X.blue : blue value                                       */
/*                                                                                         */
/*                         X: Ranging from 1 to stem2.0                                    */
/*                                                                                         */
/*                         Note: The number of entries must match                          */
/*                               stem1.bpp^2.                                              */
/*                                                                                         */
/*            stem3   in - The binary bitmap data.                                         */
/*                                                                                         */
/*            string  in - the optional filetype extension reported                        */
/*                         by GBM_Types to override autodetection                          */
/*                         based on the file extension.                                    */
/*                                                                                         */
/* Return:    "0", an error message or "ERROR" for an unknown error                        */
/*******************************************************************************************/

#define MAX(a,b)  (a>b?a:b)

ULONG GBM_FileWrite(PUCHAR name, ULONG numargs, RXSTRING args[],
                    PSZ queuename, RXSTRING *retstr)
{
  char * filename       = NULL;
  char * options        = NULL;

  char * headerStem     = NULL, * headerStemSep  = NULL;
  char * paletteStem    = NULL, * paletteStemSep = NULL;
  char * dataStem       = NULL;

  char * stemNameBuffer = NULL;
  int    headerStemLen  = 0;
  int    paletteStemLen = 0;
  int    maxNameLength  = 0;

  int    fd             = -1;
  int    filetype       = -1;
  int    flag           = 0;

  GBM    gbm            = { 0 };
  GBMFT  gbmft          = { 0 };
  GBMRGB gbmrgb[0x100]  = { 0 };
  char   value[21]      = { 0 };

  char * data           = NULL;

  GBM_ERR rc            = GBM_ERR_NOT_SUPP;

  /* prevent compiler warnings */
  args      = args;
  name      = name;
  queuename = queuename;

  BUILDRXSTRING(retstr, ERROR_RETSTR);

  if ((numargs < 5) || (numargs > 6))
  {
    return INVALID_ROUTINE;
  }

  /* extract arguments */
  filename       = args[0].strptr;
  options        = args[1].strptr;
  headerStem     = args[2].strptr;
  headerStemSep  = (headerStem[strlen(headerStem)-1] != '.') ? "." : "";
  paletteStem    = args[3].strptr;
  paletteStemSep = (paletteStem[strlen(paletteStem)-1] != '.') ? "." : "";
  dataStem       = args[4].strptr;

  headerStemLen  = strlen(headerStem)  + strlen(headerStemSep);
  paletteStemLen = strlen(paletteStem) + strlen(paletteStemSep);

  maxNameLength  = MAX(headerStemLen, paletteStemLen);

  /* allocate the name buffer for full stem names: basename + 3 digits + 2 dots + 20 subnames */
  stemNameBuffer = (char *) malloc(maxNameLength + 3 + 2 + 20 + 1);
  if (stemNameBuffer == NULL)
  {
    BUILDRXSTRING(retstr, ERROR_MEM);
    return VALID_ROUTINE;
  }

  /* get the REXX header variables */
  {
    /* header.width */
    sprintf(stemNameBuffer, "%s%swidth", headerStem, headerStemSep);
    if (! GetRexxVariable(stemNameBuffer, value, sizeof(value)-1))
    {
      free(stemNameBuffer);
      return INVALID_ROUTINE;
    }
    if (sscanf(value, "%i", &(gbm.w)) != 1)
    {
      free(stemNameBuffer);
      return INVALID_ROUTINE;
    }

    /* header.height */
    sprintf(stemNameBuffer, "%s%sheight", headerStem, headerStemSep);
    if (! GetRexxVariable(stemNameBuffer, value, sizeof(value)-1))
    {
      free(stemNameBuffer);
      return INVALID_ROUTINE;
    }
    if (sscanf(value, "%i", &(gbm.h)) != 1)
    {
      free(stemNameBuffer);
      return INVALID_ROUTINE;
    }

    /* header.bpp */
    sprintf(stemNameBuffer, "%s%sbpp", headerStem, headerStemSep);
    if (! GetRexxVariable(stemNameBuffer, value, sizeof(value)-1))
    {
      free(stemNameBuffer);
      return INVALID_ROUTINE;
    }
    if (sscanf(value, "%i", &(gbm.bpp)) != 1)
    {
      free(stemNameBuffer);
      return INVALID_ROUTINE;
    }
  }

  /* get the REXX palette variables */
  {
    int i;
    int palette_entries = 0;

    /* number of entries */
    sprintf(stemNameBuffer, "%s%s0", paletteStem, paletteStemSep);
    if (! GetRexxVariable(stemNameBuffer, value, sizeof(value)-1))
    {
      free(stemNameBuffer);
      return INVALID_ROUTINE;
    }
    if (sscanf(value, "%i", &palette_entries) != 1)
    {
      free(stemNameBuffer);
      return INVALID_ROUTINE;
    }
    if (palette_entries > sizeof(gbmrgb)/sizeof(gbmrgb[0]))
    {
      free(stemNameBuffer);
      return INVALID_ROUTINE;
    }

    /* copy palette entries */
    for (i = 0; i < palette_entries; i++)
    {
      /* red */
      sprintf(stemNameBuffer, "%s%s%i.red", paletteStem, paletteStemSep, i+1);
      if (! GetRexxVariable(stemNameBuffer, value, sizeof(value)-1))
      {
        free(stemNameBuffer);
        return INVALID_ROUTINE;
      }
      if (sscanf(value, "%i", &(gbmrgb[i].r)) != 1)
      {
        free(stemNameBuffer);
        return INVALID_ROUTINE;
      }

      /* green */
      sprintf(stemNameBuffer, "%s%s%i.green", paletteStem, paletteStemSep, i+1);
      if (! GetRexxVariable(stemNameBuffer, value, sizeof(value)-1))
      {
        free(stemNameBuffer);
        return INVALID_ROUTINE;
      }
      if (sscanf(value, "%i", &(gbmrgb[i].g)) != 1)
      {
        free(stemNameBuffer);
        return INVALID_ROUTINE;
      }

      /* blue */
      sprintf(stemNameBuffer, "%s%s%i.blue", paletteStem, paletteStemSep, i+1);
      if (! GetRexxVariable(stemNameBuffer, value, sizeof(value)-1))
      {
        free(stemNameBuffer);
        return INVALID_ROUTINE;
      }
      if (sscanf(value, "%i", &(gbmrgb[i].b)) != 1)
      {
        free(stemNameBuffer);
        return INVALID_ROUTINE;
      }
    }
  }

  /* get the REXX bitmap data */
  {
    const int stride  = ( ((gbm.w * gbm.bpp + 31)/32) * 4 );
    const int datalen = stride * gbm.h;

    /* Allocate a buffer for the bitmap data */
    data = (char *) malloc((size_t) datalen);
    if (data == NULL)
    {
      free(stemNameBuffer);
      BUILDRXSTRING(retstr, ERROR_MEM);
      return VALID_ROUTINE;
    }

    if (! GetRexxBinaryVariable(dataStem, data, datalen))
    {
      free(data);
      free(stemNameBuffer);
      return INVALID_ROUTINE;
    }

    free(stemNameBuffer);
  }

  Gbm_init();

  /* check if filetype override is provided */
  if (numargs > 5)
  {
    filetype = lookupGbmFileType(args[5].strptr);
    if (filetype < 0)
    {
      free(data);
      Gbm_deinit();
      BUILDRXSTRING(retstr, ERROR_FILETYPE);
      return VALID_ROUTINE;
    }
  }
  else
  {
    /* query filetype from file extension */
    rc = Gbm_guess_filetype(filename, &filetype);
    if (rc != GBM_ERR_OK)
    {
      free(data);
      BUILDRXSTRING(retstr, formatGbmErrorMessage(rc));
      Gbm_deinit();
      return VALID_ROUTINE;
    }
  }

  /* check if the output format supports the requested color depth */
  rc = Gbm_query_filetype(filetype, &gbmft);
  if (rc != GBM_ERR_OK)
  {
    BUILDRXSTRING(retstr, formatGbmErrorMessage(rc));
    Gbm_deinit();
    return VALID_ROUTINE;
  }

  switch(gbm.bpp)
  {
    case 64: flag = GBM_FT_W64; break;
    case 48: flag = GBM_FT_W48; break;
    case 32: flag = GBM_FT_W32; break;
    case 24: flag = GBM_FT_W24; break;
    case  8: flag = GBM_FT_W8;  break;
    case  4: flag = GBM_FT_W4;  break;
    case  1: flag = GBM_FT_W1;  break;
    default: flag = 0;          break;
  }
  if ( (gbmft.flags & flag) == 0 )
  {
    free(data);
    Gbm_deinit();
    BUILDRXSTRING(retstr, ERROR_BPP_UNSUPP);
    return VALID_ROUTINE;
  }

  /* create file */
  fd = Gbm_io_create(filename, GBM_O_WRONLY);
  if (fd != -1)
  {
    rc = Gbm_write(filename, fd, filetype, &gbm, gbmrgb, data, options);
    if (rc != GBM_ERR_OK)
    {
      BUILDRXSTRING(retstr, formatGbmErrorMessage(rc));
      free(data);
      Gbm_io_close(fd);
      remove(filename);
      Gbm_deinit();
      return VALID_ROUTINE;
    }

    free(data);
    Gbm_io_close(fd);

    /* build return string */
    BUILDRXSTRING(retstr, NOERROR_RET);

    return VALID_ROUTINE;
  }

  /* build return string */
  BUILDRXSTRING(retstr, ERROR_FILECREATE);

  Gbm_deinit();
  return VALID_ROUTINE;
}



